@@ -2,15 +2,15 @@
 #
 # miniirc - A small-ish IRC framework.
 #
-# © 2019 by luk3yx and other developers of miniirc.
+# © 2018-2020 by luk3yx and other contributors of miniirc.
 #
 
 import atexit, errno, threading, time, socket, ssl, sys
 
 # The version string and tuple
-ver = __version_info__ = (1,5,0)
-version = 'miniirc IRC framework v1.5.0'
-__version__ = '1.5.0'
+ver = __version_info__ = (1,6,0)
+version = 'miniirc IRC framework v1.6.0'
+__version__ = '1.6.0'
 
 # __all__ and _default_caps
 __all__ = ['CmdHandler', 'Handler', 'IRC']
@@ -68,7 +68,7 @@ def _tags_to_dict(tag_list, separator=';'):
             tags[tag[0]] = True
         elif len(tag) == 2:
             if '\\' in tag[1]: # Iteration is bad, only do it if required.
-                value  = ''
+                value = ''
                 escape = False
                 for char in tag[1]: # TODO: Remove this iteration.
                     if escape:
@@ -79,7 +79,7 @@ def _tags_to_dict(tag_list, separator=';'):
                     else:
                         value += char
             else:
-                value = tag[1]
+                value = tag[1] or True
             tags[tag[0]] = value
 
     return tags
@@ -105,9 +105,9 @@ def ircv3_message_parser(msg):
         if len(i) < 2:
             i.append(i[0])
         hostmask = (hostmask[0], i[0], i[1])
-        cmd      = n[1]
+        cmd = n[1]
     else:
-        cmd      = n[0]
+        cmd = n[0]
         hostmask = (cmd, cmd, cmd)
         n.insert(0, '')
 
@@ -150,17 +150,19 @@ def _dict_to_tags(tags):
 
 # A wrapper for callable logfiles
 class _Logfile:
-    __slots__ = ('_buffer', '_func')
+    __slots__ = ('_buffer', '_func', '_lock')
 
     def write(self, data):
-        self._buffer += data
-        while '\n' in self._buffer:
-            line, self._buffer = self._buffer.split('\n', 1)
-            self._func(line)
+        with self._lock:
+            self._buffer += data
+            while '\n' in self._buffer:
+                line, self._buffer = self._buffer.split('\n', 1)
+                self._func(line)
 
     def __init__(self, func):
         self._buffer = ''
         self._func = func
+        self._lock = threading.Lock()
 
 # Replace invalid RFC1459 characters with Unicode lookalikes
 def _prune_arg(arg):
@@ -170,12 +172,12 @@ def _prune_arg(arg):
 
 # Create the IRC class
 class IRC:
-    connected  = None
+    connected = None
     debug_file = sys.stdout
-    sendq      = None
-    msglen     = 512
+    sendq = None
+    msglen = 512
     _main_lock = None
-    _sasl      = False
+    _sasl = False
     _unhandled_caps = None
 
     # This will no longer be an alias in miniirc v2.0.0.
@@ -192,7 +194,7 @@ class IRC:
     def quote(self, *msg, force=None, tags=None):
         if not tags and msg and isinstance(msg[0], dict):
             tags = msg[0]
-            msg  = msg[1:]
+            msg = msg[1:]
         if self.connected or force:
             if not isinstance(tags, dict) \
                     or ('message-tags' not in self.active_caps
@@ -221,8 +223,8 @@ class IRC:
 
     def send(self, *msg, force=None, tags=None):
         if len(msg) > 1:
-            self.quote(*map(_prune_arg, msg[:-1]), ':' + msg[-1], force=force,
-                tags=tags)
+            self.quote(*tuple(map(_prune_arg, msg[:-1])) + (':' + msg[-1],),
+                force=force, tags=tags)
         else:
             self.quote(*map(_prune_arg, msg), force=force, tags=tags)
 
@@ -231,7 +233,7 @@ class IRC:
         self.quote('PRIVMSG', str(target), ':' + ' '.join(msg), tags=tags)
 
     def notice(self, target, *msg, tags=None):
-        self.quote('NOTICE',  str(target), ':' + ' '.join(msg), tags=tags)
+        self.quote('NOTICE', str(target), ':' + ' '.join(msg), tags=tags)
 
     def ctcp(self, target, *msg, reply=False, tags=None):
         m = (self.notice if reply else self.msg)
@@ -259,7 +261,7 @@ class IRC:
             self._send_lock.release()
 
         self.debug('Connecting to', self.ip, 'port', self.port)
-        addrinfo  = socket.getaddrinfo(self.ip, self.port, 0,
+        addrinfo = socket.getaddrinfo(self.ip, self.port, 0,
             socket.SOCK_STREAM)[0]
         self.sock = socket.socket(*addrinfo[:2])
         if self.ssl:
@@ -287,7 +289,7 @@ class IRC:
 
     # Disconnect from IRC.
     def disconnect(self, msg=None, *, auto_reconnect=False):
-        self.persist   = auto_reconnect and self.persist
+        self.persist = auto_reconnect and self.persist
         self.connected = None
         self.active_caps.clear()
         atexit.unregister(self.disconnect)
@@ -338,8 +340,8 @@ class IRC:
 
     # Launch handlers
     def _handle(self, cmd, hostmask, tags, args):
-        r        = False
-        cmd      = str(cmd).upper()
+        r = False
+        cmd = str(cmd).upper()
         hostmask = tuple(hostmask)
         for handlers in (_global_handlers, self.handlers):
             if cmd in handlers:
@@ -376,6 +378,8 @@ class IRC:
                         raise
                     else:
                         self._pinged = True
+                        if self.ping_timeout:
+                            self.sock.settimeout(self.ping_timeout)
                         self.quote('PING', ':miniirc-ping', force=True)
                 except socket.error as e:
                     if e.errno != errno.EWOULDBLOCK:
@@ -428,25 +432,26 @@ class IRC:
             ssl=None, ident=None, realname=None, persist=True, debug=False,
             ns_identity=None, auto_connect=True, ircv3_caps=None,
             connect_modes=None, quit_message='I grew sick and died.',
-            ping_interval=60, verify_ssl=True):
+            ping_interval=60, ping_timeout=None, verify_ssl=True):
         # Set basic variables
-        self.ip             = ip
-        self.port           = int(port)
-        self.nick           = nick
+        self.ip = ip
+        self.port = int(port)
+        self.nick = nick
         if isinstance(channels, str):
             channels = map(str.lstrip, channels.split(','))
-        self.channels       = set(channels or ())
-        self.ident          = ident    or nick
-        self.realname       = realname or nick
-        self.ssl            = ssl
-        self.persist        = persist
-        self.ircv3_caps     = set(ircv3_caps or ()) | _default_caps
-        self.active_caps    = set()
-        self.isupport       = {}
-        self.connect_modes  = connect_modes
-        self.quit_message   = quit_message
-        self.ping_interval  = ping_interval
-        self.verify_ssl     = verify_ssl
+        self.channels = set(channels or ())
+        self.ident = ident    or nick
+        self.realname = realname or nick
+        self.ssl = ssl
+        self.persist = persist
+        self.ircv3_caps = set(ircv3_caps or ()) | _default_caps
+        self.active_caps = set()
+        self.isupport = {}
+        self.connect_modes = connect_modes
+        self.quit_message = quit_message
+        self.ping_interval = ping_interval
+        self.ping_timeout = ping_timeout
+        self.verify_ssl = verify_ssl
 
         # Set the NickServ identity
         if not ns_identity or isinstance(ns_identity, str):
@@ -467,7 +472,7 @@ class IRC:
 
         # Add handlers and set the default message parser
         self.change_parser()
-        self.handlers   = {}
+        self.handlers = {}
         self._send_lock = threading.Lock()
         if ssl == None and self.port == 6697:
             self.ssl = True
@@ -494,8 +499,9 @@ def _handler(irc, hostmask, args):
         irc.debug('*** Joining channels...', irc.channels)
         irc.quote('JOIN', ','.join(irc.channels))
 
-    if irc.sendq:
+    with self._send_lock:
         sendq, irc.sendq = irc.sendq, None
+    if irc.sendq:
         for i in sendq:
             irc.quote(*i)
 
@@ -505,8 +511,10 @@ def _handler(irc, hostmask, args):
 
 @Handler('PONG', colon=False)
 def _handler(irc, hostmask, args):
-    if args and args[-1] == 'miniirc-ping':
+    if args and args[-1] == 'miniirc-ping' and irc.ping_interval:
         irc._pinged = False
+        if irc.ping_timeout:
+            irc.sock.settimeout(irc.ping_interval)
 
 @Handler('432', '433')
 def _handler(irc, hostmask, args):
@@ -517,8 +525,8 @@ def _handler(irc, hostmask, args):
             pass
         if len(irc.nick) >= irc.isupport.get('NICKLEN', 20):
             return
-        print('WARNING: The requested nickname', repr(irc.nick), 'is invalid or'
-            ' already in use. Trying again with', repr(irc.nick + '_') + '...',
+        irc.debug('WARNING: The requested nickname', repr(irc.nick), 'is '
+            'invalid. Trying again with', repr(irc.nick + '_') + '...',
             file=sys.stderr)
         irc.nick += '_'
         irc.quote('NICK', irc.nick, force=True)
@@ -543,7 +551,7 @@ def _handler(irc, hostmask, args):
     cmd = args[1].upper()
     if cmd in ('LS', 'NEW'):
         caps = args[-1].split(' ')
-        req  = set()
+        req = set()
         if not irc._unhandled_caps:
             irc._unhandled_caps = {}
         for raw in caps:
@@ -555,7 +563,9 @@ def _handler(irc, hostmask, args):
                     irc._handle_cap(cap)
                 else:
                     req.add(cap)
-        if req:
+        if irc.connected is None:
+            return
+        elif req:
             irc.quote('CAP REQ', ':' + ' '.join(req), force=True)
         elif cmd == 'LS' and not irc._unhandled_caps and args[2] != '*':
             irc._unhandled_caps = None
@@ -615,7 +625,7 @@ def _handler(irc, hostmask, args):
         irc.debug('NOTICE: An IRCv3 STS has been detected, the port will',
             'be changed to', port, 'and TLS/SSL will be enabled.')
         irc.port = port
-        irc.ssl  = True
+        irc.ssl = True
         time.sleep(1)
         irc.connect()
         irc.persist = persist
